import re
import ROOT
import zfit
import math
import numpy
import pprint
import pandas            as pnd
import utils_noroot      as utnr
import matplotlib.pyplot as plt

from logzero     import logger as log
from zutils.plot import plot   as zfp

#--------------------------------
class extractor:
    '''
    Class used to extract RK from data, model, constraints, etc
    '''
    def __init__(self, drop_correlations=False):
        '''
        Parameters
        ------------------
        drop_correlations (bool): Will diagonalize the covariance matrix used for constraints
        if set to True, default is False.
        '''
        self._l_dset      = ['all'] 
        self._drop_corr   = drop_correlations

        self._preffix     = None 
        self._ck_val      = None
        self._ck_cov      = None
        self._d_data      = None
        self._d_model     = None
        self._d_ck        = None 
        self._d_val       = None
        self._d_var       = None
        self._d_par       = None
        self._d_pref      = None
        self._l_fix       = None
        self._d_fix       = {'Name' : [], 'Value' : []} 

        self._rk          = zfit.Parameter('rk', 1.0, 0.0, 2.0)

        self._plt_dir     = None

        self._has_const   = False
        self._initialized = False
    #--------------------------------
    def _initialize(self):
        if self._initialized:
            return

        self._check_inputs()
        self._d_ck = { dsname : zfit.Parameter(f'ck_{dsname}_{self._preffix}', 1.0, 0.0, 2.0) for dsname in self._l_dset }

        self._d_par   = self._get_model_pars()
        self._d_pref  = { name : par.value().numpy() for name, par in self._d_par.items() }

        d_model = {}
        for ds, (l_mod_mm, mod_ee) in self._d_model.items():
            [mod_bm_mm, mod_qm_mm] = l_mod_mm
            mod_bm_mm = self._reparametrize_mm_pdf(ds, mod_bm_mm, mod_ee)

            d_model[ds] = [mod_bm_mm, mod_qm_mm], mod_ee

        self._d_model = d_model

        self._check_constraints()
        self._fix_pars()

        self._initialized = True
    #--------------------------------
    def _get_model_pars(self):
        d_par = {}
        for [mod_bm_mm, mod_qm_mm], mod_ee in self._d_model.values():
            d_par_bm_mm = { par.name : par for par in mod_bm_mm.get_params() }
            d_par_qm_mm = { par.name : par for par in mod_qm_mm.get_params() }
            d_par_ee    = { par.name : par for par in mod_ee.get_params()    }

            d_par.update(d_par_bm_mm)
            d_par.update(d_par_qm_mm)
            d_par.update(d_par_ee)

        return d_par
    #--------------------------------
    def _fix_pars(self):
        if self._l_fix is None:
            log.info('Not fixing any parameters')
            return

        for fix_par in self._l_fix:
            l_par_name = []
            for par_name, par in self._d_par.items():
                if fix_par in par_name:
                    l_par_name.append(par_name)
                    self._add_fixed_par(par)

            if len(l_par_name) == 0:
                log.error(f'No parameters found to fix for: {fix_par}')
                l_par_name = list(self._d_par.keys())
                log.info(l_par_name)
                raise
    #--------------------------------
    def _check_constraints(self):
        '''
        Checks that every parameter constrained, exists in the model
        '''
        if not self._has_const:
            return

        for var_name in self._d_val.keys():
            if var_name not in self._d_par:
                log.error(f'Constraint for missing model variable: {var_name}')
                pprint.pprint(l_mod_var_name)
                raise
    #--------------------------------
    def _check_inputs(self):
        for ds, ([mod_bm_mm, mod_qm_mm], mod_ee) in self._d_model.items():
            self._check_model(mod_bm_mm)
            self._check_model(mod_ee   )

        for ds, ([dat_bm_mm, dat_qm_mm], dat_ee) in self._d_data.items():
            self._check_data(dat_bm_mm)
            self._check_data(dat_qm_mm)
            self._check_data(dat_ee   )

        self._check_keys(self._d_model, self._d_data )
    #--------------------------------
    def _check_keys(self, d1, d2):
        if d1.keys() != d2.keys():
            log.error('Keys differ:')
            log.error(d1.keys())
            log.error(d2.keys())
            raise ValueError
    #--------------------------------
    def _reparametrize_mm_pdf(self, dsname, mod_mm, mod_ee):
        nsg_mm  = self._get_yld(mod_mm, 'nsg_mm')
        nsg_ee  = self._get_yld(mod_ee, 'nsg_ee')

        l_model = mod_mm.models
        s_param = mod_mm.get_params(is_yield=True)
        l_param = list(s_param)
        index   = l_param.index(nsg_mm)

        ck     = self._d_ck[dsname]
        nsg_mm = zfit.ComposedParameter(f'nsg_mm_rk_{dsname}', lambda a, b, c: a * b / c, params=[self._rk, nsg_ee, ck])

        log.debug(f'nsg_mm_rk_{dsname} -> {nsg_ee.name} * {self._rk.name} / {ck.name}')

        l_param[index] = nsg_mm

        l_model_ext  = [ model.copy().create_extended(nevt, name=model.name) for model, nevt in zip(l_model, l_param) ]

        mod_mm = zfit.pdf.SumPDF(l_model_ext)

        return mod_mm
    #--------------------------------
    def _add_fixed_par(self, par):
        par.floating          = False
        self._d_fix['Name']  += [par.name]
        self._d_fix['Value'] += [float(par.numpy())]

        log.warning(f'{par.name:<60}{"=":20}{par.value():<20.3e}')
    #--------------------------------
    def _get_yld(self, model, preffix):
        l_param   = model.get_params(is_yield=True)

        l_sig_yld = [ param for param in l_param if param.name.startswith(preffix)]
        [sig_yld] = l_sig_yld

        return sig_yld
    #--------------------------------
    @property
    def plt_dir(self):
        return self._plt_dir 

    @plt_dir.setter
    def plt_dir(self, value):
        try:
            self._plt_dir = utnr.make_dir_path(value)
        except:
            log.error(f'Cannot create: {value}')
            raise
    #--------------------------------
    @property
    def fix(self):
        return self._l_fix

    @fix.setter
    def fix(self, value):
        self._l_fix = value
    #--------------------------------
    @property
    def ck(self):
        return self._ck_val

    @ck.setter
    def ck(self, value):
        self._ck_val = value 
    #--------------------------------
    @property
    def cov(self):
        return self._ck_cov

    @cov.setter
    def cov(self, value):
        self._ck_cov = value 
    #--------------------------------
    @property
    def data(self):
        return self._d_data

    @data.setter
    def data(self, value):
        self._d_data = value 
    #--------------------------------
    @property
    def model(self):
        return self._d_model

    @model.setter
    def model(self, value):
        self._d_model = value 
    #--------------------------------
    @property
    def const(self):
        return self._d_val, self._d_var

    @const.setter
    def const(self, value):
        '''
        It will take the constraints on extra parameters. 
        All the variances that are None (in second dictionary of tuple) correspond to fixed parameters

        Parameters:
        --------------------
        tuple : Storing two dictionaries {par_name : par_value}, {par_name : par_variance}
        '''

        val, var    = value
        self._d_val = val 
        self._d_var = var

        if self._d_val.keys() != self._d_var.keys():
            log.error(f'Keys of values and variances differ')
            raise ValueError

        self._has_const = True
    #--------------------------------
    def _check_data(self, obj):
        if not isinstance(obj, zfit.data.Data):
            log.error(f'Object introduced is not a zfit dataset')
            raise ValueError
    #--------------------------------
    def _check_model(self, obj):
        if not isinstance(obj, zfit.pdf.SumPDF):
            log.error(f'Object introduced is not a zfit PDF')
            raise ValueError

        pdf = obj
        if not pdf.is_extended:
            log.error(f'PDF is not extended:')
            print(pdf)
            raise ValueError

        l_yld = pdf.get_params(is_yield=True)
        l_yld_name = [ yld.name for yld in l_yld if yld.name.startswith('nsg_') ]

        try:
            [yld_name] = l_yld_name
        except:
            log.error('Not found one and only one signal yield:')
            print(l_yld_name)
            raise ValueError

        self._check_preffix(yld_name)

        log.debug(f'Picking up component with signal yield: {yld_name}')
    #--------------------------------
    def _check_preffix(self, yld_name):
        regex= r'n[a-z]+_(ee|mm)_[a-z0-9]+_(.*)'
        mtch = re.match(regex, yld_name)
        if not mtch:
            log.error(f'Cannot match yield name: {yld_name} with {regex}')
            raise

        if   self._preffix is None:
            self._preffix = mtch.group(2)
        elif self._preffix != mtch.group(2):
            log.error(f'Found different prefixes: {self._preffix} -> {mtch.group(3)}')
            raise
    #--------------------------------
    def _get_nsig(self, model):
        s_par_flt = model.get_params(floating=True)
        s_par_fix = model.get_params(floating=False)
        s_par     = s_par_flt.union(s_par_fix)

        d_par = { par.name : par.value().numpy() for par in s_par if par.name == 'rk' or par.name.startswith('ck_') or par.name.startswith('nsg_ee_')}
        if len(d_par) == 1:
            [(nam, val)] = d_par.items()
            return val, nam
        elif len(d_par) != 3:
            log.error('Cannot find right parameters in model:')
            pprint.pprint(d_par)
            raise

        rk   = d_par['rk']
        [ck] = [ val for nam, val in d_par.items() if nam.startswith('ck_'    ) ]
        [ne] = [ val for nam, val in d_par.items() if nam.startswith('nsg_ee_') ]
        val  = ne * rk / ck

        [nam] = [ par.name for par in s_par if par.name.startswith('nsg_ee_') ]
        nam   = nam.replace('_ee_', '_mm_').replace('_TIS_', '_TOS_')

        return val, nam
    #--------------------------------
    def _get_stats(self, model, data, is_muon):
        arr_dat      = data.numpy().flatten()
        pst_fit, nam = self._get_nsig(model)
        pre_fit      = self._d_pref[nam]
        gen_val      = float(arr_dat.size)

        v1 = f'True: {pre_fit:.0f}'
        v2 = f'Genr: {gen_val:.0f}'
        v3 = f'Post: {pst_fit:.0f}'

        stats_str = f'{v1}\n{v2}\n{v3}' if  is_muon else  f'{v1}\n{v3}'

        return stats_str
    #--------------------------------
    def _get_legend(self):
        d_leg         = {}
        d_leg['prc']  = r'$c\bar{c}_{prc} + \psi(2S)K^+$'
        d_leg['bpks'] = r'$B^+\to K^{*+}e^+e^-$'
        d_leg['bdks'] = r'$B^0\to K^{*0}e^+e^-$'
        d_leg['bsph'] = r'$B_s\to \phi e^+e^-$'
        d_leg['bpk1'] = r'$B^+\to K_{1}e^+e^-$'
        d_leg['bpk2'] = r'$B^+\to K_{2}e^+e^-$'

        return d_leg
    #--------------------------------
    def _plot(self, data, model, results, component, stacked=None):
        if self._plt_dir is None:
            return

        stats = self._get_stats(model, data, is_muon = component.startswith('mm_') )

        obj=zfp(data=data, model=model, result=results)
        obj.plot(nbins=60, d_leg = self._get_legend(), stacked=stacked, ext_text=stats) 
        if   'ee'in component and 'TOS' in component:
            obj.axs[0].set_ylim(0, 120)
        elif 'mm'in component and 'TOS' in component:
            obj.axs[0].set_ylim(0, 250)
        elif 'ee'in component and 'TIS' in component:
            obj.axs[0].set_ylim(0,  70)

        obj.axs[0].grid()
        obj.axs[1].set_ylim(-5, 5)
        obj.axs[1].axhline(0, color='r')

        plt_path = f'{self._plt_dir}/fit_{component}.png'
        log.info(f'Saving to: {plt_path}')
        plt.savefig(plt_path, bbox_inches='tight')
    #--------------------------------
    def _finalize(self):
        self._delete_par('rk')

        for ck in self._d_ck.values():
            self._delete_par(ck.name)

        for dsname in self._l_dset:
            self._delete_par(f'nsg_ee_rk_{dsname}')

        df_fix=pnd.DataFrame(self._d_fix)
        df_fix.to_json(f'{self._plt_dir}/fixed_pars.json', indent=4)
    #--------------------------------
    def _delete_par(self, par_name):
        if par_name in zfit.Parameter._existing_params:
            del zfit.Parameter._existing_params[par_name]
    #--------------------------------
    def _randomize(self, l_val, cov):
        l_val_rdm = numpy.random.multivariate_normal(l_val, cov, 1)[0]

        log.debug(f'Randomizing Gaussian mean: {l_val} -> {l_val_rdm}')

        return l_val_rdm.tolist()
    #--------------------------------
    def _randomize_ck(self):
        if self._l_fix is not None and 'ck' in self._l_fix:
            log.info(f'CK was already fixed, not randomizing and fixing it')
            return

        ck_par = self._d_ck['all']
        ck_err = math.sqrt(self._ck_cov)
        ck_val = numpy.random.normal(loc=self._ck_val, scale=ck_err, size=1)
        ck_val = float(ck_val)

        log.info('Fixing CK')
        ck_par.set_value(ck_val)
        ck_par.floating = False
    #--------------------------------
    def _get_ex_const(self):
        if not self._has_const:
            log.info('No extra constraints found')
            return 

        log.info('Getting extra constraints')
        l_par    = [ self._d_par[parname] for parname in self._d_val.keys() ]
        l_val    = [ self._d_val[parname] for parname in self._d_val.keys() ]
        l_var    = [ self._d_var[parname] for parname in self._d_val.keys() ]
        cov      = numpy.diag(l_var)
        l_val    = self._randomize(l_val, cov)

        return l_par, l_val, cov 
    #--------------------------------
    def _plot_fit(self):
        if self._plt_dir is None:
            return

        for dsname in self._l_dset:
            mod_mm, mod_ee = self._d_model[dsname]
            dat_mm, dat_ee = self._d_data [dsname]

            self._plot(dat_mm, mod_mm, None, f'mm_{dsname}_stk', stacked=True)
            self._plot(dat_ee, mod_ee, None, f'ee_{dsname}_stk', stacked=True)

            self._plot(dat_mm, mod_mm, None, f'mm_{dsname}_ovr', stacked=False)
            self._plot(dat_ee, mod_ee, None, f'ee_{dsname}_ovr', stacked=False)
    #--------------------------------
    def _combine_matrices(self, mat_1, mat_2):
        rows = mat_1.shape[0] + mat_2.shape[0]
        cols = mat_1.shape[1] + mat_2.shape[1]
        
        mat_3 = numpy.zeros((rows, cols))
        
        mat_3[:mat_1.shape[0], :mat_1.shape[1]] = mat_1
        mat_3[mat_1.shape[0]:, mat_1.shape[1]:] = mat_2

        return mat_3
    #--------------------------------
    def _merge_constraints(self, ex_cns, ck_cns):
        if   ex_cns is None: 
            l_par = ck_cns[0]
            l_val = ck_cns[1]
            cov   = ck_cns[2]
        elif ck_cns is None:
            l_par = ex_cns[0]
            l_val = ex_cns[1]
            cov   = ex_cns[2]
        else:
            l_par = ex_cns[0] + ck_cns[0]
            l_val = ex_cns[1] + ck_cns[1]
            cov   = self._combine_matrices(ex_cns[2], ck_cns[2])

        if self._drop_corr:
            log.warning('Using covariance matrix with correlations removed')
            cov_no_diag=numpy.matrix.copy(cov)
            numpy.fill_diagonal(cov_no_diag, 0)
            cov = cov - cov_no_diag

        return l_par, l_val, cov
    #--------------------------------
    def _build_const(self):
        ex_cns            = self._get_ex_const()
        if ex_cns is None: 
            return

        l_par, l_val, cov = ex_cns
        cns   = zfit.constraint.GaussianConstraint(params     = l_par,
                                                   observation= l_val,
                                                   uncertainty= cov)

        self._dump_constraints(l_par, l_val, cov)

        return cns
    #--------------------------------
    def _dump_constraints(self, l_par, l_val, cov):
        npar   = len(l_par)
        l_name = [par.name           for par   in l_par]
        l_cov  = [ cov[i_par][i_par] for i_par in range(npar)] 
        l_err  = [ math.sqrt(cov)    for cov   in l_cov ]
        df     = pnd.DataFrame({'Name' : l_name, 'Value' : l_val, 'Error' : l_err})

        df.to_json(f'{self._plt_dir}/constraints.json', indent=4)
    #--------------------------------
    @utnr.timeit
    def get_fit_result(self):
        self._initialize()

        log.info('Creating likelihoods')
        nll = None 
        for dsname in self._l_dset:
            mod_mm, mod_ee = self._d_model[dsname]
            dat_mm, dat_ee = self._d_data [dsname]

            nll_mm = zfit.loss.ExtendedUnbinnedNLL(model=mod_mm, data=dat_mm)
            nll_ee = zfit.loss.ExtendedUnbinnedNLL(model=mod_ee, data=dat_ee)

            if nll is None:
                nll = nll_mm + nll_ee
            else:
                nll+= nll_mm + nll_ee

        self._randomize_ck()
        cns = self._build_const()
        if cns is not None:
            nll = nll.create_new(constraints=cns)

        mnm = zfit.minimize.Minuit()
        log.info('Running minimization')
        res = mnm.minimize(nll)

        #self._plot_fit()

        log.info('Finalizing')
        self._finalize()

        return res 
#--------------------------------





sig_name = 'Signal'

def get_bdt_bin_settings(bdt_bin = None):
    '''
    For a given BDT bin, will return corresponding settings
    Parameters
    ----------------
    bdt_bin (int): Index of bin

    Returns
    ----------------
    tuple with:

    d_bdt_wp (dict): Dictionary between BDT variable and bin boundaries, e.g. {'BDT_cmb' : [0.831, 0.900]}
    vra (str)      : Version used to scale rare PRec with respect to signal, it's BDT dependent through efficiencies
    vcc (str)      : Version used to pick up right charm PRec shape
    '''
    if   bdt_bin == 0:
        d_bdt_wp  = {'BDT_cmb' : [0.627, 0.697]}
        vra       = 'v5'
        vcc       = 'v0'
    elif bdt_bin == 1:
        d_bdt_wp  = {'BDT_cmb' : [0.697, 0.767]}
        vra       = 'v5'
        vcc       = 'v1'
    elif bdt_bin == 2:
        d_bdt_wp  = {'BDT_cmb' : [0.767, 0.837]}
        vra       = 'v5'
        vcc       = 'v2'
    elif bdt_bin == 3:
        d_bdt_wp  = {'BDT_cmb' : [0.837, 0.907]}
        vra       = 'v5'
        vcc       = 'v3'
    elif bdt_bin == 4:
        d_bdt_wp  = {'BDT_cmb' : [0.907, 0.977]}
        vra       = 'v4'
        vcc       = 'v4'
    elif bdt_bin == 5:
        d_bdt_wp  = {'BDT_cmb' : [0.977, 1.047]}
        vra       = 'v3'
        vcc       = 'v5'
    else:
        log.error(f'Invalid BDT bin: {bdt_bin}')
        raise ValueError

    d_bdt_wp['BDT_prc'] = [0.480751, 10]

    return d_bdt_wp




import os
import numpy
import pandas       as pnd
import utils_noroot as utnr

from version_management  import get_last_version 
from importlib.resources import files
from ndict               import ndict
from rk.eff_yld_loader   import eff_yld_loader as eyl
from log_store           import log_store

log = log_store.add_logger('rk_extractor:np_reader')
#----------------------------------------
class np_reader:
    '''
    Class used to read nuisance parameters to calculate RK
    '''
    #------------------------
    def __init__(self, sys=None, sta=None, yld=None):
        '''
        sys (str): Version of efficiencies obtained when assessing systematics
        sta (str): Version of efficiencies obtained when assessing statistical uncertainties with bootstrapping
        yld (str): Version of fitted data yields (only Jpsi and Psi2S)
        '''
        self._sys         = sys 
        self._sta         = sta
        self._yld         = yld

        self._cov_dir     = None
        self._df_yld      = None 
        self._df_eff      = None 

        self._sys_flg     = 'pall_tall_gall_lall_hall_rall_qall_iall_snom'
        self._sta_flg     = 'pnom_tnom_gnom_lnom_hnom_rnom_qnom_inom_sall'
        self._nom_flg     = 'pnom_tnom_gnom_lnom_hnom_rnom_qnom_inom_snom'

        self._l_ds_lab    = ['r1_TOS', 'r1_TIS', 'r2p1_TOS', 'r2p1_TIS', '2017_TOS', '2017_TIS', '2018_TOS', '2018_TIS']
        self._l_dset      = ['r1', 'r2p1', '2017', '2018']
        self._l_trig      = ['MTOS', 'ETOS']
        self._l_proc      = ['sign', 'ctrl']

        self._will_cache  = False
        self._initialized = False 
    #------------------------
    def _initialize(self):
        if self._initialized:
            return

        self._setup_paths()
        self._load_eff_yld()
        self._load_uncertainties()
        self._cache()

        self._initialized = True
    #------------------------
    def _setup_paths(self):
        cache_dir = files('extractor_data').joinpath(f'npr_data/{self._sys}_{self._sta}_{self._yld}')
        os.makedirs(cache_dir, exist_ok=True)

        self._eff_path = f'{cache_dir}/eff.json'
        self._yld_path = f'{cache_dir}/yld.json'
        self._sys_path = f'{cache_dir}/sys.json'
        self._sta_path = f'{cache_dir}/sta.json'
    #------------------------
    def _find_cached(self, path):
        if not os.path.isfile(path):
            log.debug(f'File not cached: {path}')
            return False 

        log.debug(f'File cached: {path}')

        return True
    #------------------------
    def _load_eff_yld(self):
        '''
        For a given channel, ee or mm, it will fill the dictionary of efficiencies for each
        dataset and trigger.
        '''
        is_eyl = self._find_cached(self._eff_path) and self._find_cached(self._yld_path) 
        if is_eyl:
            log.info(f'Cached data found')
            self._df_eff = pnd.read_json(self._eff_path)
            self._df_yld = pnd.read_json(self._yld_path)

            return

        self._will_cache = True
        log.info(f'Cached efficiencies and yields not found, caching')

        d_eff   = self._get_dset_dict()
        d_yld   = self._get_dset_dict()
        l_index = []
        for proc in self._l_proc: 
            for trig in self._l_trig:
                channel      = 'mm' if trig == 'MTOS' else 'ee'

                eff_1, yld_1 = self._get_eff_yld(f'{proc}_{channel}', 'r1'  , trig)
                eff_2, yld_2 = self._get_eff_yld(f'{proc}_{channel}', 'r2p1', trig)
                eff_3, yld_3 = self._get_eff_yld(f'{proc}_{channel}', '2017', trig)
                eff_4, yld_4 = self._get_eff_yld(f'{proc}_{channel}', '2018', trig)

                d_eff['r1'  ].append(eff_1)
                d_eff['r2p1'].append(eff_2)
                d_eff['2017'].append(eff_3)
                d_eff['2018'].append(eff_4)

                d_yld['r1'  ].append(yld_1)
                d_yld['r2p1'].append(yld_2)
                d_yld['2017'].append(yld_3)
                d_yld['2018'].append(yld_4)

                l_index.append(f'{proc}_{trig}')

        self._df_eff = pnd.DataFrame(d_eff, index=l_index).T
        self._df_yld = pnd.DataFrame(d_yld, index=l_index).T
    #------------------------
    def _load_uncertainties(self):
        is_cov = self._find_cached(self._sys_path) and self._find_cached(self._sta_path)
        if is_cov: 
            log.info(f'Cached covariances found, loading them')
            self._cvsys = utnr.load_json(self._sys_path)
            self._cvsta = utnr.load_json(self._sta_path)
            return

        self._will_cache = True
        eff_dir          = os.environ['EFFDIR']
        self._cov_dir    = f'{eff_dir}/../covariance'

        self._cvsys = self._get_cov(kind='sys')
        log.warning(f'Statistical uncertainy covariance was taken from v65')
        self._cvsta = self._get_cov(kind='sta')
    #------------------------
    def _cache(self):
        if not self._will_cache:
            log.debug('Data was not updated, skipping caching')
            return

        log.info(f'Data was updated, caching:')

        self._df_eff.to_json(self._eff_path, indent=4)
        self._df_yld.to_json(self._yld_path, indent=4)

        utnr.dump_json(self._cvsys, self._sys_path)
        utnr.dump_json(self._cvsta, self._sta_path)
    #------------------------
    def _get_dset_dict(self):
        d_data         = dict()
        d_data['r1'  ] = []
        d_data['r2p1'] = []
        d_data['2017'] = []
        d_data['2018'] = []

        return d_data
    #------------------------
    def _get_eff_yld(self, proc, year, trig):
        '''
        Will return numerical value of efficiency and fitted yield, for a specifi process
        in a year and trigger
        '''
        obj          = eyl(proc, trig, year, self._nom_flg)
        obj.eff_var  = 'B_PT'
        t_yld, d_eff = obj.get_values(eff_version = self._sta, yld_version=self._yld)

        ctf  = d_eff['nom', 'B_PT']
        deff = ctf.efficiency
        oeff = deff.efficiency()
        eff  = oeff.val[0]
        yld  = t_yld[0]

        return eff, yld
    #------------------------
    def _get_cov(self, kind=None):
        if kind not in ['sys', 'sta']:
            log.error(f'Invalid uncertainty: {kind}')
            raise ValueError

        eff_ver  = self._sys if kind == 'sys' else self._sta
        pkl_path = f'{self._cov_dir}/{eff_ver}_{self._yld}/rx/matrix_abs_rc/tot.pkl'
        log.info(f'Picking up covariance from: {pkl_path}')
        cov      = utnr.load_pickle(pkl_path)

        return cov.tolist()
    #------------------------
    def get_cov(self, kind=None):
        '''
        Will return covariance matrix (nxn numpy array)
        '''
        self._initialize()

        if kind not in ['sys', 'sta']:
            log.error(f'Invalid covariance kind: {kind}')
            raise ValueError

        cov_mat = self._cvsys if kind == 'sys' else self._cvsta

        return numpy.array(cov_mat)
    #------------------------
    def get_eff(self):
        '''
        Will return rare mode efficiencies

        d_eff (dict): Dictionary {ds : (eff_mm, eff_ee)} with efficiency objects
        '''
        self._initialize()

        l_eff_rare_mm = self._df_eff.sign_MTOS.to_numpy().tolist()
        l_eff_rare_ee = self._df_eff.sign_ETOS.to_numpy().tolist()

        d_eff = {}
        for ds_lab, eff_mm, eff_ee in zip(self._l_ds_lab, l_eff_rare_mm, l_eff_rare_ee):
            d_eff[ds_lab] = eff_mm, eff_ee

        return d_eff 
    #------------------------
    def get_byields(self):
        '''
        Will return dictionary with efficiency corrected yields {ds : yld}
        e.g. {'r1_TIS_ee': 40021323}
        '''
        self._initialize()

        df_yld = self._df_yld.divide(self._df_eff)
        df_yld = df_yld.drop(columns=['sign_MTOS', 'sign_ETOS'])

        return df_yld
    #------------------------
    def get_ryields(self):
        self._initialize()

        br_bpjp_jpmm = self._get_reso_br()
        br_rare      = self._get_rare_br()

        df_yld       = self.get_byields()
        df_yld       = df_yld.divide(br_bpjp_jpmm)

        #This is the number of B mesons, before any selection
        #it will be multiplied by rare BR and efficiencies, so it will be a signal dataframe
        df_yld = df_yld.rename(columns={'ctrl_MTOS' : 'sign_MTOS', 'ctrl_ETOS' : 'sign_ETOS'})
        df_yld = df_yld.multiply(br_rare)

        df_era = self._df_eff.drop(columns=['ctrl_MTOS', 'ctrl_ETOS'])
        df_yld = df_yld.multiply(df_era)

        return df_yld
    #------------------------
    def _get_sfdata(self):
        sf_dir = files('extractor_data').joinpath('rare_sf')
        vers   = get_last_version(dir_path=sf_dir, version_only=True) 
        sf_pat = f'{sf_dir}/{vers}/fr_bf.json'
        if not os.path.isfile(sf_pat):
            log.error(f'File not found: {sf_pat}')
            raise FileNotFoundError

        d_data = utnr.load_json(sf_pat)

        return d_data 
    #------------------------
    def _get_reso_br(self):
        d_data = self._get_sfdata()
        b1, _  = d_data['bf']['bpjk']
        b2, _  = d_data['bf']['jpmm']

        return b1 * b2
    #------------------------
    def _get_rare_br(self):
        d_data = self._get_sfdata()
        br, _  = d_data['bf']['bpkp']

        return br 
    #------------------------
    def get_rjpsi(self):
        '''
        Will return an array with rjpsi for every trigger and dataset 
        '''
        self._initialize()

        df_yld = self.get_byields()

        sr_rjpsi_tos = df_yld.ctrl_MTOS / df_yld.ctrl_ETOS

        df_rjpsi = pnd.DataFrame({'TOS' : sr_rjpsi_tos})

        return df_rjpsi 
#----------------------------------------




import numpy
import pprint
import rk.utilities as rkut

from np_reader import np_reader as np_rdr
from cmb_ck    import combiner  as cmb_ck
#---------------------------------------------
def test_simple():
    rdr          = np_rdr(sys='v65', sta='v63', yld='v24')
    rdr.cache    = True
    d_eff        = rdr.get_eff()
    d_byld       = rdr.get_byields()
    d_nent       = rkut.average_byields(d_byld, l_exclude=['TIS'])
    d_rare_yld   = rkut.reso_to_rare(d_nent, kind='jpsi')

    cmb                 = cmb_ck(rk=1, eff=d_eff, yld=d_rare_yld) 
    cmb.out_dir         = 'tests/cmb_eff/simple'
    t_comb              = cmb.get_combination()
    d_rjpsi, d_eff, cov = t_comb
#---------------------------------------------
if __name__ == '__main__':
    test_simple()



from rk_model  import rk_model as model
from log_store import log_store

import zutils.utils as zut
import pprint
import zfit
import re

log = log_store.add_logger('rk_extractor:model_manager')
#---------------------------------------------------------------------------
class manager:
    reparametrize=True
    #---------------------------------------------------------------------------
    def __init__(self, preffix=None, d_eff=None, d_nent=None, dset=None, chan=None):
        self._preffix = preffix
        self._d_eff   = d_eff 
        self._d_nent  = d_nent
        self._dset    = dset
        self._channel = chan

        self._obs_mm_sp  = zfit.Space('mass mm', limits=(2600, 3300))
        self._cmb_pref   = 'ncb_ee'

        self._out_dir    = None
        self._d_data     = None
        self._d_mode     = None
        self._a          = zfit.param.Parameter('a', -100,-1e5,   0)  
        self._b          = zfit.param.Parameter('b', 1000,   0, 1e5)
        self._l_bdt_bin  = [1, 2, 3, 4, 5]
        self._fake_model = False

        self._initialized = False
    #---------------------------------------------------------------------------
    def _initialize(self):
        if self._initialized:
            return

        self._set_instances()
        self._set_data_model()
        self._reparametrize_cmb_yield()
        self._print_model()

        self._initialized = True
    #---------------------------------------------------------------------------
    @property
    def bdt_bin(self):
        return self._l_bdt_bin

    @bdt_bin.setter
    def bdt_bin(self, value):
        self._l_bdt_bin = value
    #---------------------------------------------------------------------------
    @property
    def fake(self):
        return self._fake_model

    @fake.setter
    def fake(self, value):
        self._fake_model = value
    #---------------------------------------------------------------------------
    @property
    def out_dir(self):
        return self._out_dir

    @out_dir.setter
    def out_dir(self, value):
        self._out_dir = value
    #---------------------------------------------------------------------------
    def _get_comb_yield_info(self, pdf):
        s_par = pdf.get_params(is_yield=True)
        [par] = [ par for par in s_par if par.name.startswith(self._cmb_pref) ]

        name  = par.name
        try:
            [sbin]= re.match('.*_(\d)$', name).groups()
        except:
            log.error(f'Cannot extract bin index from: {name}')
            raise

        ibin = int(sbin)

        return name, ibin
    #---------------------------------------------------------------------------
    def _reparametrize_cmb_yield(self):
        if not manager.reparametrize:
            log.warning(f'Not reparametrizing combinatorial yield')
            return

        d_mode = {}
        for x, (y, ee_mod) in self._d_mode.items(): 
            if ee_mod is None:
                d_mode[x]  = y, ee_mod  
                continue

            name, ibin = self._get_comb_yield_info(ee_mod)
            d_par      = self._get_yields({f'{name}_rep' : ibin})
            ee_mod     = self._reparametrize(ee_mod, d_par)
            d_mode[x]  = y, ee_mod

        self._d_mode = d_mode
    #------------------------------------
    def _get_yields(self, d_par_in):
        d_par_ot = {}
        for name, ibin in d_par_in.items():
            d_par_ot[name] = zfit.ComposedParameter(name, lambda a, b: ibin * a + b, params={'a' : self._a, 'b' : self._b })

        return d_par_ot 
    #------------------------------------
    def _reparametrize(self, pdf, d_par):
        l_pdf   = []
        pdf_name= pdf.name
        for pdf_com in pdf.pdfs:
            nev  = pdf_com.get_yield()
            name = f'{nev.name}_rep'
            if name in d_par:
                par     = d_par[name]
                pdf_com = pdf_com.copy()
                pdf_com.set_yield(par)
    
            l_pdf.append(pdf_com)

        return zfit.pdf.SumPDF(l_pdf, name=pdf_name)
    #---------------------------------------------------------------------------
    def _set_instances(self):
        self._d_inst = {}
        for bdt_bin in self._l_bdt_bin:
            mod = model(
                    preffix  =f'{self._preffix}_{bdt_bin}', 
                    d_eff    =self._d_eff, 
                    d_nent   =self._d_nent, 
                    channel  =self._channel,
                    obs_mm_sp=self._obs_mm_sp,
                    l_dset   =[self._dset])

            mod.bdt_bin  = bdt_bin
            mod.kind     = 'nom'
            mod.out_dir  = f'{self._out_dir}/pdf/bin_{bdt_bin}'
            mod.fake     = self._fake_model 

            self._d_inst[bdt_bin] = mod
    #---------------------------------------------------------------------------
    def _get_data(self, bdt_bin):
        d_data = self._d_inst[bdt_bin].get_data()
        data   = d_data[self._dset]

        return data
    #---------------------------------------------------------------------------
    def _get_model(self, bdt_bin):
        d_mode = self._d_inst[bdt_bin].get_model()
        mode   = d_mode[self._dset]

        return mode 
    #---------------------------------------------------------------------------
    def _set_data_model(self):
        bdt_bin    = self._l_bdt_bin[-1]
        mm_data, _ = self._get_data(bdt_bin)
        mm_mode, _ = self._get_model(bdt_bin)

        d_data = {}
        d_mode = {}
        for bdt_bin in self._l_bdt_bin:
            log.debug(f'Processing BDT bin: {bdt_bin}')
            _, ee_data = self._get_data(bdt_bin)
            _, ee_mode = self._get_model(bdt_bin)

            d_data[f'bdt_{bdt_bin}'] = mm_data, ee_data
            d_mode[f'bdt_{bdt_bin}'] = mm_mode, ee_mode

        self._d_mode = d_mode
        self._d_data = d_data
    #---------------------------------------------------------------------------
    def _print_model(self):
        for _, (l_pdf_mm, pdf_ee) in self._d_mode.items():
            if l_pdf_mm is not None:
                for pdf_mm in l_pdf_mm:
                    zut.print_pdf(pdf_mm, level='debug')

            if pdf_ee is not None:
                zut.print_pdf(pdf_ee, level='debug')
    #---------------------------------------------------------------------------
    def get_data(self):
        self._initialize()

        return self._d_data
    #---------------------------------------------------------------------------
    def get_model(self):
        self._initialize()

        if self._d_mode is None:
            log.error('Dictionary of models is None')
            raise

        return self._d_mode
#---------------------------------------------------------------------------

